use std::{
	fs,
	io::{self, Write},
	path::PathBuf,
	str::FromStr,
};

use clap::Parser;
use cli::{CliArgs, FilesArgs, TagArgs, TagsArgs, UntagArgs};
use taggo::{init_db, try_database};

use anyhow::{Result, anyhow};

mod cli;

pub const TERM_RESET: &str = "\x1b[0m";
pub const TERM_BOLD: &str = "\x1b[1m";

fn main() -> anyhow::Result<()> {
	let args = CliArgs::parse();

	match args.commands {
		cli::Commands::Init => init_db().map_err(|_| anyhow!("Database already exists!")),
		cli::Commands::Tag {
			args,
			files,
			file_tags,
		} => handle_tag(args, files, file_tags),
		cli::Commands::Untag(args) => handle_untag(args),
		cli::Commands::Tags(args) => handle_tags(args),
		cli::Commands::Files(args) => handle_files(args),
	}
}

fn handle_files(args: FilesArgs) -> Result<()> {
	let conn = try_database()?;

	let mut sql = r#"SELECT path, group_concat(name) as tags FROM file_tag
			JOIN file ON file.id = file_tag.file_id
			JOIN tag ON tag.id = file_tag.tag_id"#
		.to_string();

	let mut show_tags = true;

	let mut paths: Vec<PathBuf> = vec![];
	if args.all {
		paths = fs::read_dir(".")?
			.filter_map(|f| f.ok())
			.filter(|f| f.metadata().is_ok() && f.metadata().unwrap().is_file())
			.map(|f| f.path())
			.collect();
	} else if let Some(query) = args.query {
		let path = PathBuf::from_str(&query).unwrap_or_default();
		match path.exists() {
			true => {
				sql.push_str(&format!(" WHERE path IN ('{}')", query));
				paths = vec![path];
			}
			false => {
				sql.push_str(&format!(" WHERE tag.name LIKE '%{}%'", query));
				show_tags = false;
			}
		};
	}

	let mut all_paths: Vec<String> = paths
		.iter_mut()
		.map(|f| f.to_string_lossy().replace("./", "").to_string())
		.collect();

	sql.push_str(" GROUP BY path");

	let mut stmt = conn.prepare(&sql)?;
	let tagged_files: Vec<(String, String)> = stmt
		.query_map([], |row| Ok((row.get(0)?, row.get(1)?)))?
		.flatten()
		.collect();

	let mut w = io::stdout();
	if !tagged_files.is_empty() {
		if args.all {
			writeln!(&mut w, "{TERM_BOLD}Tagged Files: {TERM_RESET}")?;
		}
		for (file, tags) in tagged_files {
			// Remove already tagged files
			if let Some(idx) = all_paths.iter().position(|v| *v == file) {
				all_paths.swap_remove(idx);
			}

			if show_tags {
				writeln!(&mut w, "{}: {}", file, tags)?;
			} else {
				writeln!(&mut w, "{}", file)?;
			}
		}
	} else {
		writeln!(&mut w, "No files registered!")?;
	}

	if args.all && !all_paths.is_empty() {
		writeln!(&mut w, "\n{TERM_BOLD}Untagged Files: {TERM_RESET}")?;
		for path in all_paths {
			let path_str = path;
			writeln!(&mut w, "{}", path_str)?;
		}
	}

	w.flush()?;

	Ok(())
}

pub fn handle_tag(args: TagArgs, files: Vec<PathBuf>, file_tags: Vec<String>) -> Result<()> {
	let mut conn = try_database()?;

	if let Some(file) = args.file {
		taggo::tag_file(&mut conn, &file, &args.tags)?;
	} else {
		// TODO: Maybe rework this at some point so its more optimized for batch insertion ?
		for file in &files {
			taggo::tag_file(&mut conn, file, &file_tags)?;
		}
	}

	Ok(())
}

pub fn handle_untag(args: UntagArgs) -> Result<()> {
	let mut conn = try_database()?;

	taggo::untag_file(&mut conn, args.file, args.tags)?;

	Ok(())
}

pub fn handle_tags(args: TagsArgs) -> Result<()> {
	let mut conn = try_database()?;

	match args.commands {
		Some(cli::TagsCommands::List) | None => {
			let mut w = io::stdout();
			let tags = taggo::list_tags(&conn)?;
			if !tags.is_empty() {
				for tag in tags {
					writeln!(&mut w, "{} - {}", tag.0, tag.1)?;
				}
			} else {
				writeln!(&mut w, "No tags registered!")?;
			}
			w.flush()?;
			Ok(())
		}
		Some(cli::TagsCommands::Add { add }) => taggo::add_tags(&mut conn, add),
		Some(cli::TagsCommands::Remove { remove }) => taggo::remove_tags(&mut conn, remove),
		Some(cli::TagsCommands::Rename { renames }) => taggo::rename_tag(&conn, renames),
	}
}
